"""
1 AP, 2 STA, multi-flow experiment

Flow Legend:

Flow 1) AP    ->  STA 1
Flow 2) AP    ->  STA 2
Flow 3) STA 1 ->  AP
Flow 4) STA 2 ->  AP

Experiment Legend:

Exp 0)  Flow 1 + FLow 2 + Flow 3 + Flow 4
Exp 1)  Flow 1
Exp 2)  Flow 2
Exp 3)  Flow 3
Exp 4)  Flow 4

10/23/14 - CRH - Updated to v1.0 WARPnet syntax
7/18/14 - CRH - Updated to v.95 WARPnet syntax

"""
import sys
import time
import wlan_exp.config as wlan_exp_config
import wlan_exp.util as wlan_exp_util
import wlan_exp.ltg as wlan_exp_ltg
import wlan_exp.log.util_hdf as hdf_util


#-------------------------------------------------------------------------
#  Global experiment variables
#

# NOTE: change these values to match your experiment setup
NETWORK           = '10.0.0.0' 
NODE_SERIAL_LIST  = ['W3-a-00000', 'W3-a-00001', 'W3-a-00002']

# Set the per-trial duration (in seconds)
TRIAL_TIME = 300

# Set the packet length of each LTG transmission
PACKET_LEN = 1400

# Set the channel for all nodes
CHANNEL = 1

# SSID of the BSS we will create
AP_SSID           = "WARP Multi-flow Exp"

# Set the transmission power in dBm
TX_POW = 5

# Set the Physical CCA Thresh [0,1023]
#PHY_CCA_THRESH = 250 #Use this line to ensure that Physical CS is enabled
PHY_CCA_THRESH = 1023 #Use this line to ensure that Physical CS is disabled

#-----------------------------------------------------------------------------
# Local Helper Utilities
#-----------------------------------------------------------------------------

def write_log_file(file_name, data_buffer):
    """Writes log data to a HDF5 file."""
    try:        
        # Get the byte log_data out of the WARPNet buffer
        data = data_buffer.get_bytes()
        
        # Write the byte Log_data to the file 
        hdf_util.log_data_to_hdf5(data, file_name, overwrite=True)
    except AttributeError as err:
        print("Error writing log file: {0}".format(err))
        
        
def reset_everything():    
    
    # Tell each STA to not automatically adjust its timebase according to received
    # beacons from the AP. This isn't strictly required, but it will keep the 
    # beginning of the logs cleaner by preventing any race condition between a
    # STA updating itself and receiving the broadcast_cmd_set_time message.
    n_sta_1.sta_configure(beacon_ts_update=False)    
    n_sta_2.sta_configure(beacon_ts_update=False)       
    
    wlan_exp_util.broadcast_cmd_set_time(0.0, network_config)     
     
    # Put each node in a known, good state
    for node in nodes:
        node.set_phy_cs_thresh(PHY_CCA_THRESH)
        node.set_tx_power(TX_POW)
        node.log_configure(log_full_payloads=False)
        node.reset_all()
    
    # Tell each STA to automatically adjust its timebase according to received
    # beacons from the AP
    n_sta_1.sta_configure(beacon_ts_update=True)    
    n_sta_2.sta_configure(beacon_ts_update=True)    
    
    # Set the network SSID
    n_ap.set_ssid(AP_SSID)
    
    # Force the association state between the AP and STA    
    n_ap.add_association(n_sta_1)
    n_ap.add_association(n_sta_2)
    
    # Check that the nodes are associated.  Otherwise, the LTGs below will fail.
    if not (n_ap.is_associated(n_sta_1) and n_ap.is_associated(n_sta_2)):
        print("\nERROR: Nodes are not associated.")
        print("    Ensure that the AP and the STA are associated.")
        sys.exit(0)
    
  
        
    
def print_and_log(string, file_name):
    f = open(file_name,'a');
    f.write(string)
    f.write("\n")
    f.close()
    print(string)
    


#-------------------------------------------------------------------------
#  Initialization
#
print("\nInitializing experiment\n")

# Create an object that describes the configuration of the host PC
network_config = wlan_exp_config.WlanExpNetworkConfiguration(network=NETWORK) 

# Create an object that describes the WARP v3 nodes that will be used in this experiment
nodes_config   = wlan_exp_config.WlanExpNodesConfiguration(network_config=network_config, 
                                                           serial_numbers=NODE_SERIAL_LIST) 

# Initialize the Nodes
#  This command will fail if either WARP v3 node does not respond
nodes = wlan_exp_util.init_nodes(nodes_config, network_config)

# Extract the AP and STA nodes from the list of initialized nodes
n_ap_l  = wlan_exp_util.filter_nodes(nodes, mac_high='AP')
n_sta_l = wlan_exp_util.filter_nodes(nodes, mac_high='STA')

# Check that we have a valid AP and two STAs
if (((len(n_ap_l) == 1) and (len(n_sta_l) == 2))):
    # Extract the two nodes from the lists for easier referencing below
    n_ap = n_ap_l[0]
    n_sta_1 = n_sta_l[0]
    n_sta_2 = n_sta_l[1]
else:
    print("ERROR: Node configurations did not match requirements of script.\n")
    print(" Ensure two nodes are ready, one using the AP design, one using the STA design\n")
    sys.exit(0)

n_ap.set_channel(CHANNEL)
n_sta_1.set_channel(CHANNEL)
n_sta_2.set_channel(CHANNEL)

#%%
#-------------------------------------------------------------------------
#  Run Experiments
#
print("\nRun Experiment:")


#for rate_idx in range(8): #Loop through all 8 supported PHY rates
for rate_idx in [3]: #Test only 1 PHY rate

    rate = wlan_exp_util.wlan_rates[rate_idx] 
    n_ap.set_tx_rate_unicast(rate, curr_assoc=True, new_assoc=True)    
    n_sta_1.set_tx_rate_unicast(rate, curr_assoc=True, new_assoc=True)    
    n_sta_2.set_tx_rate_unicast(rate, curr_assoc=True, new_assoc=True)    
    
    #for exp_idx in range(5):  
    for exp_idx in [0]:             
        print("\nExperiment {0}:".format(exp_idx))
        
        reset_everything()
        
        time.sleep(2)

        if exp_idx == 0:        
            flow_1_ltg_id  = n_ap.ltg_configure(wlan_exp_ltg.FlowConfigCBR(n_sta_1.wlan_mac_address, PACKET_LEN, 0, 0), auto_start=True)
            flow_2_ltg_id  = n_ap.ltg_configure(wlan_exp_ltg.FlowConfigCBR(n_sta_2.wlan_mac_address, PACKET_LEN, 0, 0), auto_start=True)
            flow_3_ltg_id  = n_sta_1.ltg_configure(wlan_exp_ltg.FlowConfigCBR(n_ap.wlan_mac_address, PACKET_LEN, 0, 0), auto_start=True)
            flow_4_ltg_id  = n_sta_2.ltg_configure(wlan_exp_ltg.FlowConfigCBR(n_ap.wlan_mac_address, PACKET_LEN, 0, 0), auto_start=True)                   
        elif exp_idx == 1:        
            flow_1_ltg_id  = n_ap.ltg_configure(wlan_exp_ltg.FlowConfigCBR(n_sta_1.wlan_mac_address, PACKET_LEN, 0, 0), auto_start=True)
            #flow_2_ltg_id  = n_ap.ltg_configure(wlan_exp_ltg.FlowConfigCBR(n_sta_2.wlan_mac_address, PACKET_LEN, 0, 0), auto_start=True)
            #flow_3_ltg_id  = n_sta_1.ltg_configure(wlan_exp_ltg.FlowConfigCBR(n_ap.wlan_mac_address, PACKET_LEN, 0, 0), auto_start=True)
            #flow_4_ltg_id  = n_sta_2.ltg_configure(wlan_exp_ltg.FlowConfigCBR(n_ap.wlan_mac_address, PACKET_LEN, 0, 0), auto_start=True)
        elif exp_idx == 2:        
            #flow_1_ltg_id  = n_ap.ltg_configure(wlan_exp_ltg.FlowConfigCBR(n_sta_1.wlan_mac_address, PACKET_LEN, 0, 0), auto_start=True)
            flow_2_ltg_id  = n_ap.ltg_configure(wlan_exp_ltg.FlowConfigCBR(n_sta_2.wlan_mac_address, PACKET_LEN, 0, 0), auto_start=True)
            #flow_3_ltg_id  = n_sta_1.ltg_configure(wlan_exp_ltg.FlowConfigCBR(n_ap.wlan_mac_address, PACKET_LEN, 0, 0), auto_start=True)
            #flow_4_ltg_id  = n_sta_2.ltg_configure(wlan_exp_ltg.FlowConfigCBR(n_ap.wlan_mac_address, PACKET_LEN, 0, 0), auto_start=True)
        elif exp_idx == 3:        
            #flow_1_ltg_id  = n_ap.ltg_configure(wlan_exp_ltg.FlowConfigCBR(n_sta_1.wlan_mac_address, PACKET_LEN, 0, 0), auto_start=True)
            #flow_2_ltg_id  = n_ap.ltg_configure(wlan_exp_ltg.FlowConfigCBR(n_sta_2.wlan_mac_address, PACKET_LEN, 0, 0), auto_start=True)
            flow_3_ltg_id  = n_sta_1.ltg_configure(wlan_exp_ltg.FlowConfigCBR(n_ap.wlan_mac_address, PACKET_LEN, 0, 0), auto_start=True)
            #flow_4_ltg_id  = n_sta_2.ltg_configure(wlan_exp_ltg.FlowConfigCBR(n_ap.wlan_mac_address, PACKET_LEN, 0, 0), auto_start=True)
        elif exp_idx == 4:        
            #flow_1_ltg_id  = n_ap.ltg_configure(wlan_exp_ltg.FlowConfigCBR(n_sta_1.wlan_mac_address, PACKET_LEN, 0, 0), auto_start=True)
            #flow_2_ltg_id  = n_ap.ltg_configure(wlan_exp_ltg.FlowConfigCBR(n_sta_2.wlan_mac_address, PACKET_LEN, 0, 0), auto_start=True)
            #flow_3_ltg_id  = n_sta_1.ltg_configure(wlan_exp_ltg.FlowConfigCBR(n_ap.wlan_mac_address, PACKET_LEN, 0, 0), auto_start=True)
            flow_4_ltg_id  = n_sta_2.ltg_configure(wlan_exp_ltg.FlowConfigCBR(n_ap.wlan_mac_address, PACKET_LEN, 0, 0), auto_start=True)
        
        
        # Record the initial Tx/Rx stats
        flow_1_rx_stats_start = n_sta_1.stats_get_txrx(n_ap)
        flow_2_rx_stats_start = n_sta_2.stats_get_txrx(n_ap)
        flow_3_rx_stats_start = n_ap.stats_get_txrx(n_sta_1)
        flow_4_rx_stats_start = n_ap.stats_get_txrx(n_sta_2)
        
        # Wait for the TRIAL_TIME
        time.sleep(TRIAL_TIME)
        
        # Record the ending Tx/Rx stats
        flow_1_rx_stats_end = n_sta_1.stats_get_txrx(n_ap)
        flow_2_rx_stats_end = n_sta_2.stats_get_txrx(n_ap)
        flow_3_rx_stats_end = n_ap.stats_get_txrx(n_sta_1)
        flow_4_rx_stats_end = n_ap.stats_get_txrx(n_sta_2)
        
        # Object duplication just so the naming convention is less confusing
        flow_1_tx_stats_start = flow_3_rx_stats_start;
        flow_2_tx_stats_start = flow_4_rx_stats_start;
        flow_3_tx_stats_start = flow_1_rx_stats_start;
        flow_4_tx_stats_start = flow_2_rx_stats_start;
        flow_1_tx_stats_end = flow_3_rx_stats_end;
        flow_2_tx_stats_end = flow_4_rx_stats_end;
        flow_3_tx_stats_end = flow_1_rx_stats_end;
        flow_4_tx_stats_end = flow_2_rx_stats_end;
        
        # Stop all LTG traffic at each node
        n_ap.ltg_stop_all()
        n_sta_1.ltg_stop_all()
        n_sta_2.ltg_stop_all()
        
        n_ap.ltg_remove_all()
        n_sta_1.ltg_remove_all()
        n_sta_2.ltg_remove_all()
        
        n_ap.queue_tx_data_purge_all()
        n_sta_1.queue_tx_data_purge_all()
        n_sta_2.queue_tx_data_purge_all()
        
        time.sleep(2)        
        
        write_log_file("E{0}_R{1:>.0f}_sta_1.hdf5".format(exp_idx, rate['rate']), n_sta_1.log_get_all_new(log_tail_pad=0))
        write_log_file("E{0}_R{1:>.0f}_sta_2.hdf5".format(exp_idx, rate['rate']), n_sta_2.log_get_all_new(log_tail_pad=0))
        write_log_file("E{0}_R{1:>.0f}_ap.hdf5".format(exp_idx, rate['rate']), n_ap.log_get_all_new(log_tail_pad=0))    
        
        # Compute the throughput
        print_and_log("     Rate = {0} Mbps -- Flow 1:".format(wlan_exp_util.tx_rate_to_str(rate)),
                      "E{0}_results.txt".format(exp_idx))
        flow_1_num_bytes   = float((flow_1_rx_stats_end['data_num_rx_bytes'] - flow_1_rx_stats_start['data_num_rx_bytes']) * 8)
        flow_1_time_span   = float(flow_1_rx_stats_end['timestamp'] - flow_1_rx_stats_start['timestamp'])
        flow_1_xput        = flow_1_num_bytes / flow_1_time_span       
        
        print_and_log("         Throughput = {0:>5.2f} Mbps".format(flow_1_xput),
                      "E{0}_results.txt".format(exp_idx))

        print_and_log("     Rate = {0} Mbps -- Flow 2:".format(wlan_exp_util.tx_rate_to_str(rate)),
                      "E{0}_results.txt".format(exp_idx))
        flow_2_num_bytes   = float((flow_2_rx_stats_end['data_num_rx_bytes'] - flow_2_rx_stats_start['data_num_rx_bytes']) * 8)
        flow_2_time_span   = float(flow_2_rx_stats_end['timestamp'] - flow_2_rx_stats_start['timestamp'])
        flow_2_xput        = flow_2_num_bytes / flow_2_time_span

        print_and_log("         Throughput = {0:>5.2f} Mbps".format(flow_2_xput),
                      "E{0}_results.txt".format(exp_idx))
        
        print_and_log("     Rate = {0} Mbps -- Flow 3:".format(wlan_exp_util.tx_rate_to_str(rate)),
                      "E{0}_results.txt".format(exp_idx))
        flow_3_num_bytes   = float((flow_3_rx_stats_end['data_num_rx_bytes'] - flow_3_rx_stats_start['data_num_rx_bytes']) * 8)
        flow_3_time_span   = float(flow_3_rx_stats_end['timestamp'] - flow_3_rx_stats_start['timestamp'])
        flow_3_xput        = flow_3_num_bytes / flow_3_time_span
            
        print_and_log("         Throughput = {0:>5.2f} Mbps".format(flow_3_xput),
                      "E{0}_results.txt".format(exp_idx))
        
        print_and_log("     Rate = {0} Mbps -- Flow 4:".format(wlan_exp_util.tx_rate_to_str(rate)),
                      "E{0}_results.txt".format(exp_idx))
        flow_4_num_bytes   = float((flow_4_rx_stats_end['data_num_rx_bytes'] - flow_4_rx_stats_start['data_num_rx_bytes']) * 8)
        flow_4_time_span   = float(flow_4_rx_stats_end['timestamp'] - flow_4_rx_stats_start['timestamp'])
        flow_4_xput        = flow_4_num_bytes / flow_4_time_span
 
        print_and_log("         Throughput = {0:>5.2f} Mbps".format(flow_4_xput),
                      "E{0}_results.txt".format(exp_idx))
        print_and_log("     Total DL Throughput (Flow 1 + Flow 2):  {0:>5.2f} Mbps".format(flow_1_xput + flow_2_xput),
                      "E{0}_results.txt".format(exp_idx))
        print_and_log("     Total UL Throughput (Flow 3 + Flow 4):  {0:>5.2f} Mbps".format(flow_3_xput + flow_4_xput),
                      "E{0}_results.txt".format(exp_idx))
        
        
